
"""
datazen - A child class for adding template-loading capabilities to the
          environment dataset.
"""

# built-in
from typing import Dict, List

# third-party
import jinja2

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.classes.base_environment import BaseEnvironment, LOADTYPE
from datazen.templates import load as load_templates


class TemplateEnvironment(BaseEnvironment):
    """
    The template-loading environment mixin, only requires the base environment
    capability to function.
    """

    def load_templates(
        self,
        template_loads: LOADTYPE = (None, None),
        name: str = ROOT_NAMESPACE
    ) -> Dict[str, jinja2.Template]:
        """ Load templates, resolve any un-loaded template directories. """

        # determine directories that need to be loaded
        data_type = DataType.TEMPLATE
        to_load = self.get_to_load(data_type, name)

        # load new templates
        template_data = self.get_data(data_type, name)
        if to_load:
            template_data.update(load_templates(to_load, template_loads[0],
                                                template_loads[1]))
            self.update_load_state(data_type, to_load, name)

        return template_data

    def add_template_dirs(self, dir_paths: List[str], rel_path: str = ".",
                          name: str = ROOT_NAMESPACE,
                          allow_dup: bool = False) -> int:
        """
        Add template directories, return the number of directories added.
        """

        return self.add_dirs(DataType.TEMPLATE, dir_paths, rel_path, name,
                             allow_dup)




"""
datazen - Reading and writing a known header signature to output files.
"""

# built-in
import os
from typing import List, Tuple

# internal
from datazen import VERSION, PKG_NAME
from datazen.parsing import get_hash

BARRIER = "="


def get_comment_data(file_data: str,
                     dynamic: bool = True) -> List[Tuple[str, str]]:
    """
    Get tuples (key-value pairs) of data to write into the file fingerprint.
    """

    line_data = [(False, ("generator", PKG_NAME)),
                 (True, ("version", VERSION)),
                 (False, ("hash", get_hash(file_data)))]

    # filter out any possibly undesired data
    resolved = []
    for line in line_data:
        if not line[0] or dynamic:
            resolved.append(line[1])

    return resolved


def apply_barriers(comment_lines: List[str], char: str) -> None:
    """
    Optionally add decoration to the comment lines by adding a barrier at
    the beginning and end.
    """

    barrier = BARRIER * max(int(len(com) / len(char)) for com in comment_lines)
    if comment_lines:
        comment_lines.insert(0, barrier)
        comment_lines.append(barrier)
        comment_lines.append("")


def resolve_encapsulation(comment_lines: List[str], file_ext: str) -> str:
    """
    Turn the requested line data into something that can be embedded in the
    target file type.
    """

    new_lines = []

    ext = file_ext.lower()
    if ext == "py":
        for line in comment_lines:
            new_lines.append("# " + line if line else line)
    elif ext == "md":
        new_lines.append("<!--")
        for line in comment_lines:
            if line:
                new_lines.append("    " + line)
        new_lines.append("-->" + os.linesep + os.linesep)

    return os.linesep.join(new_lines)


def build_fingerprint(file_data: str, file_ext: str,
                      char: str = BARRIER, dynamic: bool = True) -> str:
    """
    Build a String that should be prepended to the final file output for
    user awareness and automated interpretation.
    """

    comment_lines = []
    line_data = get_comment_data(file_data, dynamic)
    for line in line_data:
        comment_lines.append("{}={}".format(line[0], line[1]))
    apply_barriers(comment_lines, char)
    return resolve_encapsulation(comment_lines, file_ext)




"""
datazen - A class for exposing the capability to run concrete tasks against
          the current environment.
"""

# built-in
from collections import defaultdict
import logging
import os
from typing import List, Tuple, Dict

# internal
from datazen import ROOT_NAMESPACE
from datazen.classes.manifest_environment import set_output_dir
from datazen.classes.base_environment import get_dep_slug, dep_slug_unwrap
from datazen.classes.manifest_cache_environment import ManifestCacheEnvironment

LOG = logging.getLogger(__name__)


class TaskEnvironment(ManifestCacheEnvironment):
    """
    Adds capability for running tasks (including tasks with dependencies)
    and tracking whether or not updates are required.
    """

    def valid_noop(self, entry: dict, _: str,
                   __: dict = None,
                   ___: List[str] = None) -> Tuple[bool, bool]:
        """ Default handle for a potentially-unregistered operation. """

        LOG.error("'%s' no handler found (default: '%s')", entry["name"],
                  self.default)
        return False, False

    def __init__(self):
        """ Add a notion of 'visited' targets to the environment data. """

        super().__init__()
        self.visited = defaultdict(bool)
        self.is_new = defaultdict(bool)
        self.default = "noop"
        self.handles = defaultdict(lambda: self.valid_noop)
        self.task_data = defaultdict(lambda: defaultdict(dict))

    def is_resolved(self, operation: str, target: str) -> bool:
        """
        Determine whether a target for an operation has already been resolved.
        """

        return self.visited[get_dep_slug(operation, target)]

    def is_task_new(self, operation: str, target: str) -> bool:
        """ Determine if a target has been newly executed this iteration. """

        return self.is_new[get_dep_slug(operation, target)]

    def resolve(self, operation: str, target: str, should_cache: bool,
                namespace: str, is_new: bool) -> None:
        """ Set a target for an operation as resolved. """

        self.visited[get_dep_slug(operation, target)] = True
        self.is_new[get_dep_slug(operation, target)] = is_new
        if should_cache:
            self.write_cache()
        if namespace != ROOT_NAMESPACE:
            self.restore_cache()

    def push_deps(self, dep_dict: dict, task_stack:
                  List[Tuple[str, str]], curr_target: str) -> None:
        """ Push dependencies onto a stack if they need to be resolved. """

        for dep in dep_dict:
            dep_tup = dep_slug_unwrap(dep, self.default)
            is_new = self.is_task_new(dep_tup[0], dep_tup[1])
            is_resolved = self.is_resolved(dep_tup[0], dep_tup[1])
            if dep != curr_target and (not is_resolved or is_new):
                task_stack.append(dep_tup)

    def get_dep_data(self, dep_list: List[str]) -> dict:
        """
        From a list of dependencies, create a dictionary with any task data
        they've saved.
        """

        dep_data = {}

        # flatten all of the tasks' data into a single dict
        for dep in dep_list:
            dep_tup = dep_slug_unwrap(dep, self.default)
            curr_data = self.task_data[dep_tup[0]][dep_tup[1]]
            if isinstance(curr_data, dict):
                dep_data.update(self.task_data[dep_tup[0]][dep_tup[1]])

        return dep_data

    def already_satisfied(self, target: str, output_path: str,
                          load_deps: List[str],
                          deps_changed: List[str] = None,
                          load_checks: Dict[str, List[str]] = None) -> bool:
        """
        Check if a target is already satisfied, if not debug-log some
        information about why not.
        """

        is_file = os.path.isfile(output_path)
        newly_loaded = self.get_new_loaded(load_deps, load_checks)
        result = (not self.manifest_changed and is_file and not deps_changed
                  and newly_loaded == 0)

        # log less-obvious dependency-resolution hits
        if is_file:
            if deps_changed:
                LOG.debug("%s: %s updates detected", target, deps_changed)
            if newly_loaded != 0:
                LOG.debug("%s: %d file updates detected", target, newly_loaded)

        return result

    def handle_task(
        self,
        key_name: str,
        target: str,
        task_stack: List[Tuple[str, str]] = None,
        should_cache: bool = True
    ) -> Tuple[bool, bool]:
        """
        Handle the setup for manifest tasks, such as loading additional data
        directories and setting the correct output directory.
        """

        # fall back on default behavior if the manifest doesn't even have
        # an entry for this key
        if key_name not in self.manifest["data"]:
            return self.handles[key_name]({"name": ""}, ROOT_NAMESPACE)

        if task_stack is None:
            task_stack = []

        entries = self.manifest["data"][key_name]
        for data in entries:
            if target == data["name"]:
                # skip targets we know we've resolved
                if self.is_resolved(key_name, target):
                    return True, False

                LOG.debug("executing '%s'", get_dep_slug(key_name, target))

                # push dependencies
                dep_data = {}
                deps_changed = []
                if "dependencies" in data:
                    self.push_deps(data["dependencies"], task_stack, target)

                    # resolve dependencies
                    while task_stack:
                        task = task_stack.pop()
                        result = self.handle_task(task[0], task[1], task_stack,
                                                  False)

                        # if a dependency failed, propagate it up
                        if not result[0]:
                            return False, False
                        # otherwise if a dependency was updated (performed),
                        # capture it in a list
                        if result[1] or self.is_task_new(task[0], task[1]):
                            deps_changed.append(get_dep_slug(task[0], task[1]))

                    # provide dependency data as "flattened"
                    dep_data = self.get_dep_data(data["dependencies"])

                # resolve the output directory
                set_output_dir(data, self.manifest["dir"],
                               self.manifest["data"]["output_dir"])

                # load additional data directories if specified
                namespace = self.get_namespace(key_name, target, data)
                self.load_dirs(data, self.manifest["dir"], namespace, True)

                # make sure this namespace is valid
                if not self.get_valid(namespace):
                    LOG.info("namespace '%s' is invalid!", namespace)
                    return False, False

                # write-through to the cache when we complete an operation,
                # if it succeeded
                result = self.handles[key_name](data, namespace, dep_data,
                                                deps_changed)
                if result[0]:
                    self.resolve(key_name, target, should_cache, namespace,
                                 result[1])
                return result

        return False, False




"""
datazen - Tests for the 'compile' command interface.
"""

# module under test
from datazen.commands.compile import cmd_compile

# internal
from ..resources import (
    get_test_configs,
    get_test_schemas,
    get_test_variables,
    get_tempfile,
)


def test_compile():
    """ Test the 'compile' command entry. """

    config_dirs = get_test_configs(True)
    schema_dirs = get_test_schemas(True)
    variable_dirs = get_test_variables(True)

    yaml_out = get_tempfile(".yaml")
    assert cmd_compile(config_dirs, schema_dirs, variable_dirs, yaml_out.name)
    json_out = get_tempfile(".json")
    assert cmd_compile(config_dirs, schema_dirs, variable_dirs, json_out.name)
    toml_out = get_tempfile(".toml")
    assert not cmd_compile(config_dirs, schema_dirs, variable_dirs,
                           toml_out.name)


def test_invalid_compile():
    """ Test the off-nominal paths for the 'compile' command. """

    config_dirs = get_test_configs(False)
    schema_dirs = get_test_schemas(False)
    variable_dirs = get_test_variables(False)

    yaml_out = get_tempfile(".yaml")
    assert not cmd_compile(config_dirs, schema_dirs, variable_dirs,
                           yaml_out.name)




"""
datazen - A child class for adding variable-data loading capabilities to the
          environment dataset.
"""

# built-in
from typing import List

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.classes.base_environment import BaseEnvironment, LOADTYPE
from datazen.variables import load as load_variables


class VariableEnvironment(BaseEnvironment):
    """
    The variable-data loading environment mixin. Only requires the base
    environment capability to function.
    """

    def load_variables(self, var_loads: LOADTYPE = (None, None),
                       name: str = ROOT_NAMESPACE) -> dict:
        """ Load variable data, resolve any un-loaded variable directories. """

        # determine directories that need to be loaded
        data_type = DataType.VARIABLE
        to_load = self.get_to_load(data_type, name)

        # load new data
        variable_data = self.get_data(data_type, name)
        if to_load:
            variable_data.update(load_variables(to_load, var_loads[0],
                                                var_loads[1]))
            self.update_load_state(data_type, to_load, name)

        return variable_data

    def add_variable_dirs(self, dir_paths: List[str], rel_path: str = ".",
                          name: str = ROOT_NAMESPACE,
                          allow_dup: bool = False) -> int:
        """
        Add variable-data directories, return the number of directories added.
        """

        return self.add_dirs(DataType.VARIABLE, dir_paths, rel_path, name,
                             allow_dup)




"""
datazen - APIs for loading data from directory trees.
"""

# built-in
from collections import defaultdict
import logging
import os
from typing import Dict, List, Tuple

# internal
from datazen import GLOBAL_KEY
from datazen.paths import get_path_list, advance_dict_by_path
from datazen.paths import get_file_name
from datazen.parsing import load as load_raw_resolve
from datazen.parsing import set_file_hash

LOG = logging.getLogger(__name__)


def meld_and_resolve(full_path: str, existing_data: dict, variables: dict,
                     globals_added: bool = False,
                     expect_overwrite: bool = False) -> bool:
    """
    Meld dictionary data from a file into an existing dictionary, assume
    existing data is a template and attempt to resolve variables.
    """

    variables_root: dict = variables

    # allow directory/.{file_type} to be equivalent to directory.{file_type}
    key = get_file_name(full_path)
    if not key:
        data_dict = existing_data
    else:
        if key not in existing_data:
            existing_data[key] = {}
        data_dict = existing_data[key]
        if key in variables:
            variables = variables[key]

    # meld the data
    global_add_success: bool = False
    if globals_added:
        if GLOBAL_KEY not in variables:
            variables[GLOBAL_KEY] = variables_root[GLOBAL_KEY]
            global_add_success = True

    _, loaded = load_raw_resolve(full_path, variables, data_dict,
                                 expect_overwrite)

    if global_add_success:
        del variables[GLOBAL_KEY]

    return loaded


def load_dir(path: str, existing_data: dict, variables: dict = None,
             loaded_list: List[str] = None,
             hashes: Dict[str, dict] = None,
             expect_overwrite: bool = False) -> dict:
    """ Load a directory tree into a dictionary, optionally meld. """

    if variables is None:
        variables = {}

    if loaded_list is None:
        loaded_list = []

    if hashes is None:
        hashes = defaultdict(dict)

    root_abs = os.path.abspath(path)
    for root, _, files in os.walk(path):
        LOG.debug("loading '%s'", root)

        path_list = get_path_list(root_abs, root)
        iter_data = advance_dict_by_path(path_list, existing_data)
        variable_data = advance_dict_by_path(path_list, variables)

        # expose data globally, if it was provided
        added_globals: bool = False
        if GLOBAL_KEY not in variable_data:
            variable_data[GLOBAL_KEY] = variables
            added_globals = True
        else:
            msg = "can't add 'global' data to '%s', key was already found"
            LOG.info(msg, root)

        # extend the provided list of files that were newly loaded, or at
        # least have new content
        loaded_list.extend(load_files(files, root,
                                      (iter_data, variable_data,
                                       added_globals),
                                      hashes, expect_overwrite))

        if added_globals:
            del variable_data[GLOBAL_KEY]

    return existing_data


def load_dir_only(path: str, expect_overwrite: bool = False) -> dict:
    """
    A convenient wrapper for loading just directory data from a path without
    worrying about melding data, resolving variables, enforcing schemas, etc.
    """

    return load_dir(path, defaultdict(dict), None, None, None,
                    expect_overwrite)


def load_files(file_paths: List[str], root: str,
               meld_data: Tuple[dict, dict, bool],
               hashes: Dict[str, dict],
               expect_overwrite: bool = False) -> List[str]:
    """
    Load files into a dictionary and return a list of the files that are
    new or had hash mismatches.
    """

    new_or_changed = []

    # load (or meld) data
    for name in file_paths:
        full_path = os.path.join(root, name)
        assert os.path.isabs(full_path)
        if (meld_and_resolve(full_path, meld_data[0], meld_data[1],
                             meld_data[2], expect_overwrite) and
                set_file_hash(hashes, full_path)):
            new_or_changed.append(full_path)

    return new_or_changed




"""
datazen - An environment extension that exposes compilation capabilities.
"""

# built-in
import logging
from typing import List, Tuple

# internal
from datazen.classes.task_environment import TaskEnvironment
from datazen.compile import str_compile, get_compile_output
from datazen.paths import advance_dict_by_path

LOG = logging.getLogger(__name__)


class CompileEnvironment(TaskEnvironment):
    """ Leverages a cache-equipped environment to perform compilations. """

    def __init__(self):
        """
        Add compiled data to a dictionary whenever a compilation is performed.
        """

        super().__init__()
        self.handles["compiles"] = self.valid_compile

    def valid_compile(self, entry: dict, namespace: str,
                      dep_data: dict = None,
                      deps_changed: List[str] = None) -> Tuple[bool, bool]:
        """ Perform the compilation specified by the entry. """

        path, output_type = get_compile_output(entry)

        # load configs early to update cache
        data = self.cached_load_configs(namespace)

        # update this dict with the dependency data
        if dep_data is not None:
            data.update(dep_data)

        # advance the dict if it was requested
        if "index_path" in entry:
            data = advance_dict_by_path(entry["index_path"].split("."), data)

        # set task-data early, in case we don't need to re-compile
        self.task_data["compiles"][entry["name"]] = data

        # make sure this compilation needs to be performed
        if self.already_satisfied(entry["name"], path,
                                  ["configs", "variables", "schemas"],
                                  deps_changed):
            LOG.info("compile '%s' satisfied, skipping", entry["name"])
            return True, False

        mode = "a" if "append" in entry and entry["append"] else "w"
        with open(path, mode) as out_file:
            out_file.write(str_compile(data, output_type))
            LOG.info("compiled '%s' data to '%s'", output_type, path)

        return True, True




"""
datazen - A class for storing metadata about files that have been loaded.
"""

# built-in
from copy import deepcopy
from collections import defaultdict
import logging
import os
import shutil
import time
from typing import Dict, List, Tuple

# internal
from datazen import VERSION
from datazen.parsing import set_file_hash, merge, dedup_dict_lists
from datazen.load import load_dir_only
from datazen.compile import str_compile

LOG = logging.getLogger(__name__)

DATA_DEFAULT = {"hashes": defaultdict(lambda: defaultdict(dict)),
                "loaded": defaultdict(list),
                "meta": {"version": VERSION}}


class FileInfoCache:
    """ Provides storage for file hashes and lists that have been loaded. """

    def __init__(self, cache_dir: str = None):
        """ Construct an empty cache or optionally load from a directory. """

        self.data: dict = deepcopy(DATA_DEFAULT)
        self.removed_data: Dict[str, List[str]] = defaultdict(list)
        self.cache_dir: str = ""
        if cache_dir is not None:
            self.load(cache_dir)

    def load(self, cache_dir: str) -> None:
        """ Load data from a directory. """

        assert self.cache_dir == ""
        self.cache_dir = cache_dir
        os.makedirs(self.cache_dir, exist_ok=True)

        # reject things that don't belong by updating instead of assigning
        new_data = sync_cache_data(load_dir_only(self.cache_dir, True),
                                   self.removed_data)
        for key in DATA_DEFAULT:
            self.data[key].update(new_data[key])

    def get_hashes(self, sub_dir: str) -> Dict[str, dict]:
        """ Get the cached, dictionary of file hashes for a certain key. """

        return self.data["hashes"][sub_dir]

    def check_hit(self, sub_dir: str, path: str,
                  also_cache: bool = True) -> bool:
        """
        Determine if a given file already exists with its current hash in the
        cache, if not return False and optionally add it to the cache. """

        abs_path = os.path.abspath(path)
        is_new = set_file_hash(self.get_hashes(sub_dir), abs_path, also_cache)
        if also_cache and is_new:
            # guard against a failure in the "new" detection
            loaded_data = self.get_loaded(sub_dir)
            if abs_path not in loaded_data:
                loaded_data.append(abs_path)

        return not is_new

    def get_loaded(self, sub_dir: str) -> List[str]:
        """ Get the cached, list of loaded files for a certain key. """

        return self.data["loaded"][sub_dir]

    def describe(self) -> None:
        """ Describe this cache's contents for debugging purposes. """

        curr_dir = os.getcwd()
        for hash_set, hash_data in self.data["hashes"].items():
            misses = []
            total = len(self.removed_data[hash_set])
            for filename, hash_item in hash_data.items():
                total += 1
                if curr_dir in filename:
                    filename = filename[len(curr_dir) + 1:]
                if not self.check_hit(hash_set, filename, False):
                    misses.append((filename, hash_item))
                else:
                    LOG.debug("%s (%s) matched", filename,
                              time_str(hash_item["time"]))

            # log all misses / updates
            for miss in misses:
                LOG.info("%s (%s) updated", miss[0], time_str(miss[1]["time"]))
            for removed in self.removed_data[hash_set]:
                LOG.info("%s no longer present", removed)

            # log a summary
            LOG.info("%s: %d/%d match", hash_set,
                     total - (len(misses) + len(self.removed_data[hash_set])),
                     total)

        # describe metadata
        if "meta" in self.data:
            LOG.info("version: %s (current: %s)",
                     self.data["meta"]["version"], VERSION)

    def get_data(self, name: str) -> Tuple[List[str], Dict[str, dict]]:
        """ Get the tuple version of cached data. """

        return (self.get_loaded(name), self.get_hashes(name))

    def clean(self) -> None:
        """ Remove cached data from the file-system. """

        self.data = deepcopy(DATA_DEFAULT)
        if self.cache_dir != "":
            shutil.rmtree(self.cache_dir)
            os.makedirs(self.cache_dir, exist_ok=True)
            LOG.info("cleaning cache at '%s'", self.cache_dir)

    def write(self, out_type: str = "json") -> None:
        """ Commit cached data to the file-system. """

        if self.cache_dir != "":
            data = sync_cache_data(self.data, self.removed_data)
            for key, val in data.items():
                key_path = os.path.join(self.cache_dir,
                                        "{}.{}".format(key, out_type))
                with open(key_path, "w") as key_file:
                    key_file.write(str_compile(val, out_type))
            LOG.debug("wrote cache to '%s'", self.cache_dir)


def copy(cache: FileInfoCache) -> FileInfoCache:
    """ Copy one cache into a new one. """

    new_cache = FileInfoCache()

    # copy the cache
    new_cache.cache_dir = cache.cache_dir
    new_cache.data = deepcopy(cache.data)
    new_cache.removed_data = deepcopy(cache.removed_data)

    return new_cache


def meld(cache_a: FileInfoCache, cache_b: FileInfoCache) -> None:
    """ Promote all updates from cache_b into cache_a. """

    merge(cache_a.data, cache_b.data, expect_overwrite=True)


def time_str(time_s: float) -> str:
    """ Concert a timestamp to a String. """

    return time.strftime("%c", time.localtime(time_s))


def cmp_loaded_count(cache_a: FileInfoCache, cache_b: FileInfoCache,
                     name: str) -> int:
    """
    Compute the total difference in file counts (for a named group)
    between two caches.
    """

    a_loaded = cache_a.get_loaded(name)
    b_loaded = cache_b.get_loaded(name)

    # get counts of all the files from both caches
    count_data: dict = defaultdict(lambda: {"a": 0, "b": 0})
    for item in a_loaded:
        count_data[item]["a"] = a_loaded.count(item)
    for item in b_loaded:
        count_data[item]["b"] = b_loaded.count(item)

    # accumulated the differences in counts
    result = 0
    for data in count_data.values():
        result += abs(data["a"] - data["b"])
    return result


def cmp_loaded_count_from_set(cache_a: FileInfoCache, cache_b: FileInfoCache,
                              name: str, files: List[str]) -> int:
    """
    Count the number of files uniquely loaded to one cache but not the other.
    """

    result = 0
    a_loaded = cache_a.get_loaded(name)
    b_loaded = cache_b.get_loaded(name)
    for filename in files:
        full_name = os.path.abspath(filename)
        if a_loaded.count(full_name) > b_loaded.count(full_name):
            result += 1
    return result


def cmp_total_loaded(cache_a: FileInfoCache, cache_b: FileInfoCache,
                     known_types: List[str],
                     load_checks: Dict[str, List[str]] = None) -> int:
    """
    Compute the total difference in file counts for a provided set of named
    groups.
    """

    result = 0
    for known in known_types:
        if load_checks is not None and known in load_checks:
            iter_result = cmp_loaded_count_from_set(cache_a, cache_b, known,
                                                    load_checks[known])
            if iter_result > 0:
                LOG.info("%d changes detected for '%s' in '%s'", iter_result,
                         known, load_checks[known])
            result += iter_result
        else:
            iter_result = cmp_loaded_count(cache_a, cache_b, known)
            if iter_result > 0:
                LOG.info("%d changes detected for '%s'", iter_result, known)
            result += iter_result
    return result


def sync_cache_data(cache_data: dict,
                    removed_data: Dict[str, List[str]]) -> dict:
    """
    Before writing a cache to disk we want to de-duplicate items in the loaded
    list and remove hash data for files that were removed so that if they
    come back at the same hash, it's not considered already loaded.
    """

    data = dedup_dict_lists(deepcopy(cache_data))
    data["hashes"] = remove_missing_hashed_files(data["hashes"], removed_data)
    data["loaded"] = remove_missing_loaded_files(data["loaded"])
    return data


def remove_missing_hashed_files(data: dict,
                                removed_data: Dict[str, List[str]]) -> dict:
    """ Assign new hash data based on the files that are still present. """

    for category in data.keys():
        new_data = {}
        for filename in data[category].keys():
            if os.path.isfile(filename):
                new_data[filename] = data[category][filename]
            elif filename not in removed_data[category]:
                removed_data[category].append(filename)
        data[category] = new_data

    return data


def remove_missing_loaded_files(data: dict) -> dict:
    """
    Audit list elements in a dictionary recursively, assume the data is String
    and the elements are filenames, assign a new list for all of the elements
    that can be located.
    """

    for key, item in data.items():
        new_list = []
        for filename in item:
            if os.path.isfile(filename):
                new_list.append(filename)
        data[key] = new_list

    return data




"""
datazen - This package's command-line entry-point application.
"""

# built-in
import argparse

# internal
from datazen.classes.environment import from_manifest
from datazen import DEFAULT_MANIFEST


def entry(args: argparse.Namespace) -> int:
    """ Execute the requested task. """

    result = 0
    env = from_manifest(args.manifest)
    if env.get_valid():
        # clean, if requested
        if args.sync:
            env.write_cache()
        if args.clean:
            env.clean_cache()
        elif args.describe:
            env.describe_cache()
        else:
            # execute targets
            result = int(not env.execute_targets(args.targets))
    else:
        result = 1
    return result


def add_app_args(parser: argparse.ArgumentParser) -> None:
    """ Add application-specific arguments to the command-line parser. """

    parser.add_argument("-m", "--manifest", default=DEFAULT_MANIFEST,
                        help=("manifest to execute tasks from (default: " +
                              "'%(default)s')"))
    parser.add_argument("-c", "--clean", action="store_true",
                        help="clean the manifest's cache and exit")
    parser.add_argument("--sync", action="store_true",
                        help=("sync the manifest's cache (write-through) " +
                              "with the state of the file system before " +
                              "execution"))
    parser.add_argument("-d", "--describe", action="store_true",
                        help="describe the manifest's cache and exit")
    parser.add_argument("targets", nargs="*", help="target(s) to execute")




"""
datazen - Tests for the 'Environment' class.
"""

# built-in
import os

# module under test
from datazen.classes.base_environment import DataType
from datazen.classes.environment import from_manifest

# internal
from ..environment import EnvironmentMock
from ..resources import get_test_configs, get_resource


def test_environment():
    """ Basic tests for the Environment class, wrapped by EnvironmentMock. """

    env = EnvironmentMock()
    assert env

    # load configs twice, it should also load variables
    assert env.get_configs(True)
    assert env.get_configs(True)
    assert env.get_variables(True)

    # make sure duplicates are rejected
    for config_dir in get_test_configs(True):
        assert not env.valid.add_dir(DataType.CONFIG, config_dir)

    # make sure non-directories are rejected
    assert not env.valid.add_dir(DataType.CONFIG, "/this/dir/shouldn't/exist")


def test_bad_include():
    """
    Test the behavior of loading manifests when the includes have problems.
    """

    # load an invalid manifest (bad include)
    manifest_path = get_resource(os.path.join("manifests", "bad_include.yaml"),
                                 False)
    env = from_manifest(manifest_path)
    assert not env.get_valid()


def test_environment_from_manifest():
    """ Test scenarios for loading manifest files. """

    # load an invalid manifest (bad schema)
    manifest_path = get_resource(os.path.join("manifests", "test.yaml"), False)
    env = from_manifest(manifest_path)
    assert not env.get_valid()

    # make sure we can't double-load the manifest
    assert not env.load_manifest_with_cache(manifest_path)

    # load an invalid manifest (bad content)
    manifest_path = get_resource(os.path.join("manifests", "invalid.yaml"),
                                 False)
    env = from_manifest(manifest_path)
    assert not env.get_valid()
    assert env.compile("a") == (False, False)


def test_operations():
    """ Test some 'compile' and 'render' scenarios. """

    manifest = get_resource("manifest.yaml", True)
    env = from_manifest(manifest)
    env.clean_cache()
    env = from_manifest(manifest)

    assert env.manifest_changed

    assert env.render("a") == (True, True)
    assert env.render("z") == (False, False)

    # here 'b' render is already satisfied, because a compile depended on it
    assert env.render("b") == (True, False)

    assert env.render("c") == (False, False)
    assert env.render("d") == (False, False)

    assert env.compile("a") == (True, False)
    assert env.compile("a") == (True, False)
    assert env.compile("b") == (True, False)
    assert env.compile("c") == (True, False)
    assert env.compile("d") == (False, False)
    assert env.compile("f") == (True, True)
    assert env.compile("g") == (False, False)
    assert env.compile("z") == (True, True)

    env = from_manifest(manifest)
    assert not env.manifest_changed
    assert env.compile("e") == (True, False)
    assert env.render("e") == (False, False)

    # test default goals
    assert env.execute() == (True, False)
    del env.manifest["data"]["default_target"]
    assert env.execute() == (True, False)

    env.clean_cache()
    assert env.execute("groups-test") == (True, True)

    # clean the cache so that we don't commit it to the repository, it's not
    # worth the cost of using relative paths over absolute paths
    env.clean_cache()


def test_load_manifest():
    """ Test a nominal manifest-loading scenario. """

    manifest = get_resource("manifest.yaml", True)
    env = from_manifest(manifest)
    env.clean_cache()
    env = from_manifest(manifest)

    cfg_data1 = env.cached_load_configs()
    assert cfg_data1
    assert env.cached_load_configs()
    env.write_cache()

    assert env.cached_enforce_schemas(cfg_data1)

    # clean the cache
    env.clean_cache()

    cfg_data2 = env.cached_load_configs()
    assert cfg_data2
    assert cfg_data2 == cfg_data1
    env.write_cache()

    # make sure configs loaded correctly
    assert "yaml2" in cfg_data2 and "json2" in cfg_data2
    assert len(cfg_data2["top_list"]) == 6

    # clean the cache so that we don't commit it to the repository, it's not
    # worth the cost of using relative paths over absolute paths
    env.clean_cache()




"""
datazen - A data structure for runtime data.
"""

# built-in
from copy import copy, deepcopy
import logging
import os
from typing import List

# internal
from datazen.paths import resolve_dir
from datazen.enums import DataType

LOG = logging.getLogger(__name__)


class EnvironmentNamespace:
    """ The base class for management of the environment data structure. """

    def __init__(self):
        """ Base constructor. """

        self.directories = {
            DataType.CONFIG: [],
            DataType.SCHEMA: [],
            DataType.TEMPLATE: [],
            DataType.VARIABLE: [],
        }

        self.data = {
            DataType.CONFIG: {},
            DataType.SCHEMA: {},
            DataType.TEMPLATE: {},
            DataType.VARIABLE: {},
        }

        self.valid = True

    def get_to_load(self, dir_type: DataType) -> List[str]:
        """
        Build a list of the yet-to-be-loaded directories for a given data
        type.
        """

        dir_data = self.directories[dir_type]

        to_load = []
        for dir_inst in dir_data:
            if not dir_inst["loaded"]:
                to_load.append(dir_inst["path"])
        return to_load

    def unload(self, dir_type: DataType) -> None:
        """ Mark all directories for a given type as un-loaded. """

        for dir_inst in self.directories[dir_type]:
            dir_inst["loaded"] = False

    def unload_all(self) -> None:
        """ Mark all directories as unloaded. """

        for key in self.data:
            self.unload(key)

    def update_load_state(self, dir_type: DataType, to_load: List[str]) -> int:
        """
        Update the load states of directories in 'to_load' for a given
        data type.
        """

        dir_data = self.directories[dir_type]
        loaded = 0

        for dir_inst in dir_data:
            if dir_inst["path"] in to_load:
                LOG.debug("loaded '%s' as '%s'", dir_inst["path"],
                          dir_type.value)
                dir_inst["loaded"] = True
                loaded = loaded + 1

        LOG.debug("loaded %d new directories for '%s'", loaded, dir_type.value)

        return loaded

    def add_dir(self, dir_type: DataType, dir_path: str,
                rel_path: str = ".", allow_dup: bool = False) -> bool:
        """ Add a directory to be loaded for a given data type. """

        dir_list = self.directories[dir_type]
        dir_path = resolve_dir(dir_path, rel_path)

        for dir_data in dir_list:
            if dir_path == dir_data["path"]:
                def noop(*_, **__):
                    pass
                log_fn = noop if allow_dup else noop
                log_fn("not adding duplicate directory '%s' to '%s'", dir_path,
                       dir_type.value)
                return allow_dup

        if not os.path.isdir(dir_path):
            LOG.error("'%s' isn't a directory, not adding to '%s'", dir_path,
                      dir_type.value)
            return False

        dir_list.append({"path": dir_path, "loaded": False})
        LOG.debug("added '%s' to '%s'", dir_path, dir_type.value)
        return True

    def add_dirs(self, dir_type: DataType, dir_paths: List[str],
                 rel_path: str = ".", allow_dup: bool = False) -> int:
        """
        Add multiple directories for a given data type, return the number of
        directories added.
        """

        dirs_added = 0
        for dir_path in dir_paths:
            if self.add_dir(dir_type, dir_path, rel_path, allow_dup):
                dirs_added = dirs_added + 1

        if dirs_added != len(dir_paths):
            LOG.error("only loaded %d / %d directories, marking invalid",
                      dirs_added, len(dir_paths))
            self.valid = False

        return dirs_added


def clone(env: EnvironmentNamespace,
          update: EnvironmentNamespace) -> EnvironmentNamespace:
    """ Create a clone (deep copy) of an existing Environment. """

    # all we need to do is copy all of the attributes
    update.directories = deepcopy(env.directories)
    update.data[DataType.CONFIG] = deepcopy(env.data[DataType.CONFIG])
    update.data[DataType.SCHEMA] = deepcopy(env.data[DataType.SCHEMA])
    update.data[DataType.VARIABLE] = deepcopy(env.data[DataType.VARIABLE])
    update.data[DataType.TEMPLATE] = copy(env.data[DataType.TEMPLATE])
    update.valid = env.valid

    return update




"""
datazen - Test the program's entry-point.
"""

# built-in
import os

# module under test
from datazen import PKG_NAME
from datazen.entry import main as datazen_main

# internal
from .resources import get_resource


def test_entry():
    """ Test some basic command-line argument scenarios. """

    manifest = get_resource("manifest.yaml", True)
    manifest_dir = os.path.dirname(manifest)
    args = [PKG_NAME, "-m", manifest]

    assert datazen_main(args + ["--not-an-option", "asdf"]) != 0

    assert datazen_main(args) == 0
    assert datazen_main(args + ["a", "b", "c"]) == 0
    assert datazen_main(args + ["not_a_target"]) != 0
    assert datazen_main([PKG_NAME, "-C", manifest_dir, "a", "b", "c"]) == 0
    assert datazen_main(args + ["--sync", "-d"]) == 0

    # change a file to test describe's miss detection
    data_file = get_resource(os.path.join("configs", "a.yaml"), True)
    with open(data_file) as manifest_file:
        manifest_data = manifest_file.read()
    with open(data_file, "w") as manifest_file:
        manifest_file.write("a: {{a}}")

    assert datazen_main(args + ["-d"]) == 0

    # remove a file to additionally test describe's miss detection
    os.remove(data_file)
    assert datazen_main(args + ["-d"]) == 0

    # restore the changed file
    with open(data_file, "w") as manifest_file:
        manifest_data = manifest_file.write(manifest_data)

    assert datazen_main(args + ["-c"]) == 0

    # test inability to load a manifest
    assert datazen_main(args + ["-m", "not-a-manifest"]) != 0




"""
datazen - Tests for the 'render' command interface.
"""

# module under test
from datazen.commands.render import cmd_render
from datazen.commands.compile import str_compile

# internal
from .. import ENV
from ..resources import get_test_templates, get_tempfile


def test_render():
    """ Test the 'render' command entry. """

    # compile monolithic yaml
    config_out = get_tempfile(".yaml")
    with open(config_out.name, "w") as config_file:
        config_file.write(str_compile(ENV.get_configs(True), "yaml"))

    # render yaml from yaml configs
    assert cmd_render(get_test_templates(True), "a", config_out.name,
                      get_tempfile(".yaml").name)

    # compile monolithic json
    config_out = get_tempfile(".json")
    with open(config_out.name, "w") as config_file:
        config_file.write(str_compile(ENV.get_configs(True), "json"))

    # render json from json configs
    assert cmd_render(get_test_templates(True), "a", config_out.name,
                      get_tempfile(".json").name)




"""
datazen - An interface for retrieving and interacting with test data.
"""

# built-in
import os
from tempfile import NamedTemporaryFile
from typing import List, IO
import pkg_resources


def get_resource(resource_name: str, valid: bool) -> str:
    """ Locate the path to a test resource. """

    valid_str = "valid" if valid else "invalid"
    resource_path = os.path.join("data", valid_str, resource_name)
    return pkg_resources.resource_filename(__name__, resource_path)


def get_test_configs(valid: bool = True) -> List[str]:
    """ Locate test-configuration-data root. """

    return [get_resource("configs", valid)]


def get_test_schemas(valid: bool = True) -> List[str]:
    """ Locate test-schema root. """

    return [get_resource("schemas", valid)]


def get_test_templates(valid: bool = True) -> List[str]:
    """ Locate test-template root. """

    return [get_resource("templates", valid)]


def get_test_variables(valid: bool = True) -> List[str]:
    """ Locate test-variables root. """

    return [get_resource("variables", valid)]


def get_tempfile(extension: str) -> IO:
    """ Obtain a temporary file name with a specific extension. """

    return NamedTemporaryFile(suffix=extension)




"""
datazen - A centralized store for runtime data.
"""

# built-in
from collections import defaultdict
import logging
from typing import List, Tuple

# internal
from datazen.classes.base_environment import dep_slug_unwrap
from datazen.classes.compile_environment import CompileEnvironment
from datazen.classes.group_environment import GroupEnvironment
from datazen.classes.render_environment import RenderEnvironment

LOG = logging.getLogger(__name__)


class Environment(CompileEnvironment, RenderEnvironment, GroupEnvironment):
    """ A wrapper for inheriting all environment-loading capabilities. """

    def __init__(self):
        """ Add a notion of 'visited' targets to the environment data. """

        super().__init__()
        self.visited = defaultdict(bool)
        self.default = "compiles"

    def execute(self, target: str = "",
                should_cache: bool = True) -> Tuple[bool, bool]:
        """ Execute an arbitrary target. """

        # resolve a default target if one wasn't provided
        data = self.manifest["data"]
        if not target:
            if "default_target" in data:
                target = data["default_target"]
                LOG.info("using default target '%s'", target)
            elif self.default in data and data[self.default]:
                target = data[self.default][0]["name"]
                LOG.info("resolving first target '%s'", target)

        slug = dep_slug_unwrap(target, self.default)
        return self.handle_task(slug[0], slug[1], should_cache=should_cache)

    def execute_targets(self, targets: List[str]) -> bool:
        """
        Execute a list of named targets and return whether or not the entire
        set was successful.
        """

        # use a resolved, default target if none are specified
        if not targets and self.default_target():
            targets.append(self.default_target())

        for target in targets:
            task_result = self.execute(target, False)
            if not task_result[0]:
                LOG.error("target '%s' failed", target)
                return False
            if not task_result[1]:
                LOG.info("'%s' already satisfied", target)

        # write the cache at the end, if we were totally successful
        self.write_cache()
        return True

    def compile(self, target: str) -> Tuple[bool, bool]:
        """ Execute a named 'compile' target from the manifest. """

        return self.handle_task("compiles", target)

    def render(self, target: str) -> Tuple[bool, bool]:
        """ Execute a named 'render' target from the manifest. """

        return self.handle_task("renders", target)


def from_manifest(manifest_path: str) -> Environment:
    """ Load an environment object from a schema definition on disk. """

    env = Environment()

    # load the manifest
    if not env.load_manifest_with_cache(manifest_path):
        LOG.error("couldn't load manifest at '%s'", manifest_path)

    return env




"""
datazen - Top-level APIs for loading and interacting with templates.
"""

# built-in
import os
from typing import Dict, List

# third-party
import jinja2

# internal
from datazen.paths import get_file_name, get_file_ext
from datazen.parsing import set_file_hash


def update_cache_primitives(dir_path: str, loaded_list: List[str],
                            hashes: Dict[str, dict]) -> None:
    """
    From a directory path, update the 'loaded_list' and 'hashes' primitives
    that belong to an upstream cache.
    """

    for path in os.listdir(dir_path):
        fpath = os.path.abspath(os.path.join(dir_path, path))
        if os.path.isfile(fpath):
            if set_file_hash(hashes, fpath):
                loaded_list.append(fpath)


def load(template_dirs: List[str], loaded_list: List[str] = None,
         hashes: Dict[str, dict] = None) -> Dict[str, jinja2.Template]:
    """
    Load jinja2 templates from a list of directories where templates can be
    found.
    """

    result = {}

    # setup jinja environment
    loader = jinja2.FileSystemLoader(template_dirs, followlinks=True)
    env = jinja2.Environment(loader=loader, trim_blocks=True,
                             lstrip_blocks=True)

    # manually inspect directories to write into the cache
    if hashes is not None and loaded_list is not None:
        for template_dir in template_dirs:
            update_cache_primitives(template_dir, loaded_list, hashes)

    # load templates into a dictionary
    for template in env.list_templates():
        key = get_file_name(template)
        assert get_file_ext(template) == "j2"
        result[key] = env.get_template(template)
    return result




"""
datazen - Top-level APIs for loading and interacting with variables.
"""

# built-in
from typing import List, Dict

# internal
from datazen.load import load_dir


def load(directories: List[str],
         loaded_list: List[str] = None,
         hashes: Dict[str, dict] = None) -> dict:
    """ Load variable data from a list of directories. """

    result: dict = {}
    for directory in directories:
        load_dir(directory, result, None, loaded_list, hashes)
    return result



# =====================================
# generator=datazen
# version=1.1.0
# hash=5c7e9660ab8e4c2d1c03184fdda5f9f6
# =====================================

"""
datazen - Package definition for distribution.
"""

# third-party
from vmklib.setup import setup  # type: ignore

# internal
from datazen import PKG_NAME, VERSION, DESCRIPTION


author_info = {"name": "Vaughn Kottler",
               "email": "vaughnkottler@gmail.com",
               "username": "vkottler"}
pkg_info = {"name": PKG_NAME, "version": VERSION, "description": DESCRIPTION}
setup(
    pkg_info,
    author_info,
    entry_override="dz",
)




"""
datazen - An environment extension that exposes rendering capabilities.
"""

# built-in
import logging
import os
from typing import List, Tuple

# third-party
import jinja2

# internal
from datazen.classes.task_environment import TaskEnvironment
from datazen.fingerprinting import build_fingerprint
from datazen.paths import get_file_ext

LOG = logging.getLogger(__name__)


class RenderEnvironment(TaskEnvironment):
    """ Leverages a cache-equipped environment to render templates. """

    def __init__(self):
        """ Add the 'renders' handle. """

        super().__init__()
        self.handles["renders"] = self.valid_render

    def perform_render(self, template: jinja2.Template,
                       path: str, entry: dict,
                       data: dict = None) -> Tuple[bool, bool]:
        """
        Render a template to the requested path using the provided data.
        """

        try:
            render_str = template.render(data).rstrip()
            render_str += os.linesep

            # determine if the caller wanted a dynamic fingerprint or not
            dynamic = True
            if ("no_dynamic_fingerprint" in entry and
                    entry["no_dynamic_fingerprint"]):
                dynamic = False

            fprint = build_fingerprint(render_str, get_file_ext(path),
                                       dynamic=dynamic)
            render_str = fprint + render_str
            with open(path, "w") as render_out:
                render_out.write(render_str)

            # save the output into a dict for consistency
            self.task_data["renders"][entry["name"]] = render_str
        except jinja2.exceptions.TemplateError as exc:
            LOG.error("couldn't render '%s' to '%s': %s",
                      entry["name"], path, exc)
            return False, False

        LOG.info("(%s) rendered '%s'", entry["name"], path)

        return True, True

    def valid_render(self, entry: dict, namespace: str, dep_data: dict = None,
                     deps_changed: List[str] = None) -> Tuple[bool, bool]:
        """ Perform the render specified by the entry. """

        # determine the output that will be produced
        path = entry["name"]
        if "output_path" in entry:
            path = entry["output_path"]
        path = os.path.join(entry["output_dir"], path)

        # load templates
        templates = self.cached_load_templates(namespace)

        if entry["name"] not in templates:
            LOG.error("no template for key '%s' found, options: %s",
                      entry["name"], list(templates.keys()))
            return False, False

        template = templates[entry["name"]]

        # if dependencies aren't specified, use config data (but don't allow
        # an implicit 'compile')
        change_criteria = ["templates"]
        if not dep_data and "dependencies" not in entry:
            dep_data = self.cached_load_configs(namespace)
            change_criteria.append("configs")
            LOG.debug("no dependencies loaded for '%s', using config data",
                      entry["name"])

        # determine if we need to perform this render
        assert template.filename is not None
        load_checks = {"templates": [template.filename]}
        if self.already_satisfied(entry["name"], path, change_criteria,
                                  deps_changed, load_checks):
            LOG.debug("render '%s' satisfied, skipping", entry["name"])
            return True, False

        return self.perform_render(template, path, entry, dep_data)



# =====================================
# generator=datazen
# version=1.1.0
# hash=4f3a065715a84d8205d216935ad1e939
# =====================================

"""
datazen - This package's command-line entry-point (boilerplate).
"""

# built-in
import argparse
import logging
import os
import sys
from typing import List

# internal
from datazen import VERSION, DESCRIPTION
from datazen.app import entry, add_app_args


def main(argv: List[str] = None) -> int:
    """ Program entry-point. """

    result = 0

    # fall back on command-line arguments
    command_args = sys.argv
    if argv is not None:
        command_args = argv

    # initialize argument parsing
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("--version", action="version",
                        version="%(prog)s {0}".format(VERSION))
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="set to increase logging verbosity")
    parser.add_argument("-C", "--dir", default=os.getcwd(), dest="dir",
                        help=("execute from a specific directory (default: " +
                              "'%(default)s')"))
    starting_dir = os.getcwd()

    add_app_args(parser)

    # parse arguments and execute the requested command
    try:
        args = parser.parse_args(command_args[1:])
        args.version = VERSION

        # initialize logging
        log_level = logging.DEBUG if args.verbose else logging.INFO
        logging.basicConfig(level=log_level,
                            format=("%(name)-36s - %(levelname)-6s - "
                                    "%(message)s"))

        # change to the specified directory
        os.chdir(args.dir)

        # run the application
        result = entry(args)
    except SystemExit as exc:
        result = 1
        if exc.code is not None:
            result = exc.code

    # return to starting dir
    os.chdir(starting_dir)

    return result




"""
datazen - A module for various package enumerations.
"""

# built-in
from enum import Enum


class DataType(Enum):
    """ The discrete types of information that can be loaded. """

    CONFIG = "config"
    SCHEMA = "schema"
    TEMPLATE = "template"
    VARIABLE = "variable"




"""
datazen - A command implementation for compiling configuration data and
          variables into a single, monolithic output.
"""

# built-in
import logging
from typing import List

# internal
from datazen.classes.environment import Environment
from datazen.parsing import get_file_ext
from datazen.compile import str_compile

LOG = logging.getLogger(__name__)


def cmd_compile(config_dirs: List[str], schema_dirs: List[str],
                variable_dirs: List[str], output_file_path: str) -> bool:
    """
    Load configuration data by resolving variables and validating provided
    schemas. Write the result to the specified output file, inferring the type
    by the file extension.
    """

    env = Environment()

    # add directories
    env.add_config_dirs(config_dirs)
    env.add_schema_dirs(schema_dirs)
    env.add_variable_dirs(variable_dirs)

    # load data (resolves variables and enforces schemas)
    configs = env.load_configs()
    if not env.configs_valid:
        return False

    # get extension
    ext = get_file_ext(output_file_path)

    # serialize the data
    result = str_compile(configs, ext)
    if not result:
        LOG.error("can't write '%s' data to '%s', no output", ext,
                  output_file_path)
        return False

    # write the output
    with open(output_file_path, "w") as output:
        output.write(result)
    LOG.info("wrote '%s' configuration data to '%s'", ext, output_file_path)

    return True
